"""Runs in 3.4 seconds for 100 genes, so roughly a 3.4 x 180 = 9 minutes for the whole H1."""

from typing import Literal

import numpy as np
from numba import njit

from illico.utils.groups import GroupContainer
from illico.utils.math import chunk_and_fortranize, compute_pval, dense_fold_change
from illico.utils.ranking import _accumulate_group_ranksums_from_argsort


# TODO: check if njit this or not: on my mbp, it is 2 faster when not jitted
@njit(nogil=True, fastmath=True, cache=False)
def dense_ovr_mwu_kernel_over_contiguous_col_chunk(
    X: np.ndarray,
    chunk_lb: int,
    chunk_ub: int,
    grpc: GroupContainer,
    is_log1p: bool,
    use_continuity: bool = True,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
) -> tuple[np.ndarray]:
    """Compute OVR ranksum test on a dense matrix of expression counts.

    Args:
        X (np.ndarray): Input dense raw counts matrix
        grpc (GroupContainer): GroupContainer
        use_continuity (bool, optional): Apply continuity factor or not. Defaults to True.
        is_log1p (bool, optional): User-indicated flag telling if data underwent log1p
        transformation or not. Defaults to False.
        alternative (Literal["two-sided", "less", "greater"]): Type of alternative hypothesis.

    Returns:
        tuple[np.ndarray]: Two-sided p-values, U-statistic and fold change.
        Each np.ndarray of shape (n_groups, n_genes)

    Author: Rémy Dubois
    """
    if chunk_lb < 0 or chunk_ub > X.shape[1] or chunk_lb > chunk_ub:
        raise ValueError((chunk_lb, chunk_ub))

    # Convert to F-order for faster column access and sorting later
    chunk = chunk_and_fortranize(X, chunk_lb, chunk_ub, None)

    # Get ranks and tie sums
    tie_sum = np.empty(chunk.shape[1], dtype=np.float64)
    ranksums = np.zeros(shape=(grpc.counts.size, chunk.shape[1]), dtype=np.float64)
    for j in range(chunk.shape[1]):
        idxs = np.argsort(chunk[:, j])
        col_tie_sum = _accumulate_group_ranksums_from_argsort(chunk[:, j], idxs, grpc.encoded_groups, ranksums[:, j])
        tie_sum[j] = col_tie_sum

    # Compute U stats
    n = chunk.shape[0]
    n_ref = np.expand_dims(n - grpc.counts, -1)  # (g, 1)
    n_tgt = np.expand_dims(grpc.counts, -1)  # (g, 1)
    statistics = (n_ref * n_tgt + n_tgt * (n_tgt + 1) / 2) - ranksums
    mu = n_ref * n_tgt / 2.0
    # Compute pvals
    pvals = np.empty(shape=(grpc.counts.size, chunk.shape[1]), dtype=np.float64)
    for j in range(chunk.shape[1]):
        for k in range(grpc.counts.size):
            pvals[k, j] = compute_pval(
                n_ref=n_ref[k, 0],
                n_tgt=n_tgt[k, 0],
                n=n,
                tie_sum=tie_sum[j],
                U=statistics[k, j],
                mu=mu[k, 0],
                contin_corr=0.5 if use_continuity else 0.0,
                alternative=alternative,
            )

    # Get fold change
    fold_change = dense_fold_change(chunk, grpc=grpc, is_log1p=is_log1p)

    return pvals, statistics, fold_change



import math
import warnings
from typing import Literal

import numpy as np
from numba import njit
from scipy import sparse as sc_sparse

from illico.utils.groups import GroupContainer


@njit(fastmath=True, nogil=True, cache=False)
def _add_at_scalar(a: np.ndarray, b: np.ndarray, c: float | int) -> None:
    """Equivalent on np.add.at with a scalar value to accumulate.

    Args:
        a (np.ndarray): Placeholder result array
        b (np.ndarray): Indices
        c (float | int): Scalar value to accumulate in `a` at `indices`.

    Author: Rémy Dubois
    """
    for i in range(len(b)):
        a[b[i]] += c


@njit(fastmath=True, nogil=True, cache=False)
def _add_at_vec(a: np.ndarray, b: np.ndarray, c: float | int) -> None:
    """Equivalent of np.add.at with a vector holding values to accumulate.

    Args:
        a (np.ndarray): Placeholder result array
        b (np.ndarray): Indices
        c (float | int): Vector holding value to accumulate in `a` at `indices`.

    Author: Rémy Dubois
    """
    for i in range(len(b)):
        a[b[i]] += c[i]


# TODO: overload np.diff properly
@njit(fastmath=True, nogil=True, cache=False)
def diff(x: np.ndarray) -> np.ndarray:
    """Equivalent of np.diff.

    For some reasons, np.ndiff failed to compile properly.

    Args:
        x (np.ndarray): Input 1-d array

    Returns:
        np.ndarray: Results diff array, of size x.size - 1.

    Author: Rémy Dubois
    """
    assert x.ndim == 1
    result = np.empty(x.size - 1, dtype=x.dtype)
    for i in range(x.size - 1):
        result[i] = x[i + 1] - x[i]
    return result


@njit(nogil=True, fastmath=False, cache=False)
def compute_pval(
    n_ref: int,
    n_tgt: int,
    n: int,
    tie_sum: float,
    U: float,
    mu: float,
    contin_corr: float = 0.0,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
) -> float:
    """Compute p-value.

    This small piece of code was isolated here because it was duplicated in the
    all six routines.

    Args:
        n_ref (int): Number of reference (control) values (cells)
        n_tgt (int): Number of perturbed (targeted) values (cells)
        n (int): Total number of values (cells)
        tie_sum (float): Tie sum
        U (float): U-statistic
        mu (float): Mean
        contin_corr (float, optional): Continuity correction factor. Defaults to 0.0.
        alternative (Literal["two-sided", "less", "greater"]): Type of alternative hypothesis.

    Returns:
        float: P-value

    Author: Rémy Dubois
    """
    tie_corr = 1.0 - tie_sum / (n * (n - 1) * (n + 1))
    if tie_corr > 1.0e-9:  # TODO: do that properly
        sigma = np.sqrt(n_ref * n_tgt * (n_ref + n_tgt + 1) / 12.0 * tie_corr)

        if alternative == "two-sided":  # two-sided
            # Compute both-sided statistic
            U = min(U, n_ref * n_tgt - U)
            delta = U - mu
            z = (np.abs(delta) + np.sign(delta) * contin_corr) / sigma
            return math.erfc(z / math.sqrt(2.0))
        elif alternative == "greater":  # greater (right-tailed)
            delta = U - mu
            z = (delta - contin_corr) / sigma
            # P(Z > z) = 0.5 * erfc(z / sqrt(2))
            return 0.5 * math.erfc(z / math.sqrt(2.0))
        elif alternative == "less":  # less (left-tailed)
            delta = U - mu
            z = (delta + contin_corr) / sigma
            # P(Z < z) = 0.5 * erfc(-z / sqrt(2))
            return 0.5 * math.erfc(-z / math.sqrt(2.0))
        else:
            raise ValueError(f"Unsupported alternative hypothesis: {alternative}")
    else:
        return 1.0


@njit(nogil=True, cache=False)
def sampled_max(data: np.ndarray, sample_size: int = 200_000) -> float:
    max_val = -np.inf
    n = data.size
    step = max(1, n // sample_size)
    for i in range(0, n, step):
        if data[i] > max_val:
            max_val = data[i]
    return max_val


def _warn_log1p(X: np.ndarray | sc_sparse.spmatrix, is_log1p: bool, sample_size: int = 2e5):
    """Warns if user's indication and data values don't match.

    Args:
        X (np.ndarray|sc_sparse.spmatrix): Values
        is_log1p (bool): User-indicated flag
        sample_size (int, optional): Number of values to sample to estimate max. Defaults to 2e5.

    Raises:
        ValueError: If data is neither sparse nor dense.

    Author: Rémy Dubois
    """
    if isinstance(X, sc_sparse.spmatrix):
        data = X.data
    elif isinstance(X, np.ndarray):
        data = X.ravel()
    else:
        raise ValueError(f"Unsupported data type: {type(X)}")
    max_val = sampled_max(data, sample_size=sample_size)
    if is_log1p:
        if max_val > 15:
            warnings.warn(
                f"User indicated is_log1p=True, but estimated data max value is {max_val:.2f}, "
                "Which seems inconsistent. Make sure data is indeed log1p transformed.",
                UserWarning,
            )
    else:
        if max_val < 15:
            warnings.warn(
                f"User indicated is_log1p=False, but estimated data max value is {max_val:.2f}, "
                "Which seems inconsistent. Make sure data is indeed raw counts.",
                UserWarning,
            )


@njit(nogil=True, fastmath=True, cache=False)
def fold_change_from_summed_expr(group_agg_counts: np.ndarray, grpc: GroupContainer) -> np.ndarray:
    """Compute fold change from summed expression values, per group.

    Args:
        group_agg_counts (np.ndarray): Sum of expression values of shape (n_groups, n_genes)
        grpc (GroupContainer): GroupContainer holding group information

    Returns:
        np.ndarray: Fold change values of shape (n_groups, n_genes)

    Author: Rémy Dubois
    """
    assert group_agg_counts.shape[0] == grpc.counts.size
    assert group_agg_counts.ndim == 2
    mu_tgt = group_agg_counts / np.expand_dims(grpc.counts, -1)
    if grpc.encoded_ref_group == -1:
        # If one-versus-rest, the reference is all but the group
        ref_agg_counts = group_agg_counts.sum(axis=0)[np.newaxis, :] - group_agg_counts  # (n_groups, n_genes)
        ref_counts = np.expand_dims(grpc.counts.sum() - grpc.counts, -1)  # (n_groups, 1)
        mu_ref = ref_agg_counts / ref_counts
    else:
        # Else, the reference is the reference group
        mu_ref = np.expand_dims(mu_tgt[grpc.encoded_ref_group], 0)  # (1, n_genes)
    fold_change = np.where(mu_ref == 0, np.inf, mu_tgt / mu_ref)
    return fold_change


@njit(nogil=True, fastmath=True, cache=False)
def dense_fold_change(X: np.ndarray, grpc: GroupContainer, is_log1p: bool) -> np.ndarray:
    """Compute fold change from a dense array of expression counts.

    Args:
        X (np.ndarray): Expression counts
        grpc (GroupContainer): GroupContainer holding group information
        is_log1p (bool): User-indicated flag if data is log1p or not.

    Returns:
        np.ndarray: Fold change values of shape (n_groups, n_genes)

    Author: Rémy Dubois
    """
    group_agg_counts = np.zeros(shape=(grpc.counts.size, X.shape[1]), dtype=np.float64)
    # Sum expressions per group
    _add_at_vec(group_agg_counts, grpc.encoded_groups, np.expm1(X) if is_log1p else X)
    # for group_id in range(grpc.counts.size):
    #     idx_start = grpc.indptr[group_id]
    #     idx_end = grpc.indptr[group_id + 1]
    #     if is_log1p:
    #         group_agg_counts[group_id, :] = np.expm1(X[idx_start:idx_end]).sum(axis=0)
    #     else:
    #         group_agg_counts[group_id, :] = X[idx_start:idx_end].sum(axis=0)
    fold_change = fold_change_from_summed_expr(group_agg_counts, grpc)
    return fold_change


def compute_sparsity(X: np.ndarray | sc_sparse.spmatrix) -> float:
    """Compute sparsity of the data matrix.

    Args:
        X (np.ndarray | sc_sparse.spmatrix): Data matrix

    Returns:
        float: Sparsity (fraction of zero elements)

    Author: Rémy Dubois
    """
    if isinstance(X, sc_sparse.spmatrix):
        n_elements = X.shape[0] * X.shape[1]
        n_nonzero = X.nnz
    elif isinstance(X, np.ndarray):
        n_elements = X.size
        n_nonzero = np.count_nonzero(X)
    else:
        raise ValueError(f"Unsupported data type: {type(X)}")
    sparsity = 1.0 - (n_nonzero / n_elements)
    return sparsity


@njit(nogil=True, fastmath=True, cache=True, boundscheck=False)
def chunk_and_fortranize(X: np.ndarray, chunk_lb: int, chunk_ub: int, indices: np.ndarray | None = None) -> np.ndarray:
    """Vertically chunk the input array and converts it to Fortran-contiguous.

    The reason to be of the conversion is that later operations access the columns of this array so F order is advantageous. Also,
    this function performs one memory allocation instead of 2, which happens if calling np.asfortranarray on top of fancy-indexing.

    NB: If indices is None, then all rows are taken as is.

    Args:
        X (np.ndarray): Input dense array
        chunk_lb (int): Lower bound of the vertical slicing
        chunk_ub (int): Upper bound of the vertical slicing
        indices (np.ndarray): Indices to reorder rows. There can be less indices than rows in X.

    Returns:
        np.ndarray: Chunked Fortran-contiguous array with reordered rows.

    Author: Rémy Dubois
    """
    # Now just fill it by making groups contiguous vertically, this will speed up sorting later on.
    if indices is not None:
        chunk = np.empty((chunk_ub - chunk_lb, indices.size), dtype=X.dtype).T  # transpose it to get Fortran order
        for i in range(indices.size):
            for j in range(0, chunk_ub - chunk_lb):
                chunk[i, j] = X[indices[i], chunk_lb + j]
    else:
        chunk = np.empty((chunk_ub - chunk_lb, X.shape[0]), dtype=X.dtype).T  # transpose it to get Fortran order
        for i in range(X.shape[0]):
            for j in range(0, chunk_ub - chunk_lb):
                chunk[i, j] = X[i, chunk_lb + j]
    return chunk



import numpy as np
from numba import njit

from illico.utils.sparse.csc import _assert_is_csc
from illico.utils.type import CSCMatrix


@njit(nogil=True, cache=False, fastmath=True)
def _accumulate_group_ranksums_from_argsort(
    arr: np.ndarray, idx: np.ndarray, groups: np.ndarray, ranksums: np.ndarray
) -> np.ndarray:
    """
    From a given array of values, indices of sorted values (result of np.argsort) and groups,
    accumulate group rank sums in the placegolder `ranksums`.

    Args:
        arr (np.ndarray): Array of non sorted values
        idx (np.ndarray): Array of indices of sorted values
        groups (np.ndarray): Array of group indicator
        ranksums (np.ndarray): Plaeholder of shape (n_groups, ) where to accumulate rank sums

    Returns:
        np.ndarray: tie sums

    Author: Rémy Dubois
    """
    # if ranks is None:
    #     ranks = np.empty(arr.size, dtype=np.float64)
    n = idx.size
    i = 0
    tie_sum = 0.0
    while i < n:
        # find tie block
        j = i + 1
        while j < n and arr[idx[j]] == arr[idx[i]]:
            j += 1
        # average rank for indices i..j-1 (1-based)
        avg_rank = 0.5 * (i + 1 + j)  # mean of [i+1, ..., j]
        for k in range(i, j):
            if groups is not None:
                ranksums[groups[idx[k]]] += avg_rank
            else:
                # If no group is specified, ranks are sum reduced into one value for the whole column
                ranksums += avg_rank
        tie_count = j - i
        # Tie sum is the same for all froups
        tie_sum += tie_count**3 - tie_count
        i = j

    return tie_sum


@njit(nogil=True)
def rank_sum_and_ties_from_sorted(A: np.ndarray, B: np.ndarray) -> tuple[np.ndarray]:
    """Compute rank sums and tie sums from two 1-d sorted arrays.

    This routine is similar to the leetcode "merge two sorted arrays", except it
    never returns to sorted array, instead it accumulate rank sums of the second array
    and tie sums for the combined arrays.

    This routine sits at the core of the one-versus-one (or one-versus-control) asymptotic
    wilcoxon rank sum test as it allows to sort controls only once.
    Args:
        A (np.ndarray): The first sorted array (controls)
        B (np.ndarray): The second sorted array (perturbed)

    Returns:
        tuple[np.ndarray]: Ranks sum from the second array, and tie sums for the combined
        arrays.

    Author: Rémy Dubois
    """
    nA = len(A)
    nB = len(B)

    i = 0
    j = 0
    k = 0  # number of items processed so far (0-based)

    sum_ranks_B = 0.0
    tie_sum = 0.0

    # main sweep
    while i < nA and j < nB:
        # pick the smallest current value
        if A[i] < B[j]:
            v = A[i]
        else:
            v = B[j]

        # count occurrences in A
        tA = 0
        ii = i
        while ii < nA and A[ii] == v:
            tA += 1
            ii += 1

        # count occurrences in B
        tB = 0
        jj = j
        while jj < nB and B[jj] == v:
            tB += 1
            jj += 1

        t = tA + tB
        avg_rank = k + 0.5 * (t + 1)

        if t > 1:
            tie_sum += t * t * t - t

        sum_ranks_B += tB * avg_rank

        k += t
        i = ii
        j = jj

    # Drain remaining A
    while i < nA:
        v = A[i]

        # count in A
        tA = 0
        ii = i
        while ii < nA and A[ii] == v:
            tA += 1
            ii += 1

        t = tA
        avg_rank = k + 0.5 * (t + 1)

        # no B contribution because tB=0
        if t > 1:
            tie_sum += t * t * t - t

        k += t
        i = ii

    # Drain remaining B
    while j < nB:
        v = B[j]

        # count in B
        tB = 0
        jj = j
        while jj < nB and B[jj] == v:
            tB += 1
            jj += 1

        t = tB
        avg_rank = k + 0.5 * (t + 1)

        sum_ranks_B += tB * avg_rank
        if t > 1:
            tie_sum += t * t * t - t

        k += t
        j = jj

    return sum_ranks_B, tie_sum


@njit(nogil=True, cache=False)
def _sort_csc_columns_inplace(csc_matrix: CSCMatrix) -> None:
    """Sort CSC columns in place.

    Args:
        csc_matrix (CSCMatrix): Input CSC matrix.

    Author: Rémy Dubois
    """
    _assert_is_csc(csc_matrix)
    for j in range(csc_matrix.shape[1]):
        csc_matrix.data[csc_matrix.indptr[j] : csc_matrix.indptr[j + 1]].sort()


@njit(nogil=True, cache=False)
def sort_along_axis(X: np.ndarray, axis: int = 0) -> np.ndarray:
    """Sort a dense array along a given axis.

    Args:
        X (np.ndarray): Input dense array.
        axis (int, optional): Axis along which to sort. Defaults to 0.

    Returns:
        np.ndarray: Sorted array.

    Author: Rémy Dubois
    """
    sorted_X = np.empty_like(X)
    if axis == 0:
        for j in range(X.shape[1]):
            sorted_X[:, j] = np.sort(X[:, j])
    elif axis == 1:
        for i in range(X.shape[0]):
            sorted_X[i, :] = np.sort(X[i, :])
    else:
        raise ValueError(f"Axis {axis} is not supported.")
    return sorted_X


@njit(nogil=True, cache=False)
def _sort_along_axis_inplace(X: np.ndarray, axis: int = 0) -> np.ndarray:
    """Sort a dense array along a given axis.

    Args:
        X (np.ndarray): Input dense array.
        axis (int, optional): Axis along which to sort. Defaults to 0.

    Returns:
        np.ndarray: Sorted array.

    Author: Rémy Dubois
    """
    if axis == 0:
        for j in range(X.shape[1]):
            X[:, j].sort()
    elif axis == 1:
        for i in range(X.shape[0]):
            X[i, :].sort()
    else:
        raise ValueError(f"Axis {axis} is not supported.")


@njit(nogil=True, cache=False)
def check_if_sorted(arr: np.ndarray) -> bool:
    """Check if an array is sorted. O(n)

    Parameters
    ----------
    arr : np.ndarray
        1-d array to check

    Returns
    -------
    bool
        If sorted or not.

    Author: Rémy Dubois
    """
    for i in range(1, arr.size):
        if arr[i] < arr[i - 1]:
            return False
    return True


@njit(nogil=True, cache=False)
def check_indices_sorted_per_parcel(
    indices: np.ndarray,
    indptr: np.ndarray,
) -> bool:
    """Check if indices of a sparse array are sorted.

    This is esssential if input data is CSR. Indeed, chunking makes use of
    binary search on indices, which requires sorted indices.

    Parameters
    ----------
    indices : np.ndarray
        Indices
    indptr : np.ndarray
        Indptr

    Returns
    -------
    bool
        True if all indices subarrays are sorted. False otherwise.
    """
    for k in range(indptr.size - 1):
        start = indptr[k]
        end = indptr[k + 1]
        indices_slice = indices[start:end]
        if not check_if_sorted(indices_slice):
            return False
    return True



import anndata as ad
import numpy as np
from loguru import logger
from scipy import sparse

from illico.utils.groups import GroupContainer


def _log_memory_usage(adata: ad.AnnData, grpc: GroupContainer, batch_size: int, n_threads: int) -> None:
    """Log estimated memory usage of the whole routine.

    This function simply computes estimated memory footprint of the DE genes routine.

    Parameters
    ----------
    adata : ad.AnnData
        Input data
    grpc : GroupContainer
        GroupContainer
    batch_size : int
        Batch size used for processing genes
    n_threads : int
        Number of threads used for processing genes

    Author: Rémy Dubois
    """
    X = adata.X
    # Results hold fold change, p-values, statistics
    results_fp = grpc.counts.size * adata.n_vars * 3 * np.float64().nbytes
    if isinstance(X, np.ndarray):
        proc_fp = batch_size / X.shape[1] * n_threads * X.nbytes
    elif isinstance(X, sparse.spmatrix):
        proc_fp = batch_size / X.shape[1] * n_threads * (X.data.nbytes + X.indptr.nbytes + X.indices.nbytes)
    total_fp = results_fp + proc_fp
    logger.trace(
        f"Estimated RAM footprint: {total_fp / (1000**3):.2f} GB. (This includes {results_fp / (1000**3):.3f} GB for results and {proc_fp / (1000**3):.3f} GB for processing data.)"
    )



from collections import namedtuple
from typing import Any

import numpy as np

GroupContainer = namedtuple(
    "GroupContainer",
    [
        "encoded_groups",
        "counts",
        "indices",
        "indptr",
        "encoded_ref_group",
    ],
)


def encode_and_count_groups(groups: np.ndarray, ref_group: Any) -> GroupContainer:
    """Build the GroupContainer holding all group-related information.

    GroupContainer holds:
    - original group information
    - reference group (control)
    - encoded groups
    - unique raw groups
    - counts (of cell, per group)
    - indices, indptr in a RLE format
    - encoded reference group (control)

    Args:
        groups (np.ndarray): 1-d array holding group labels, one per cell
        ref_group (Any): Flag

    Returns:
        GroupContainer: GroupContainer holding all group-related information.

    Author: Rémy Dubois
    """
    if ref_group not in groups and ref_group is not None:
        raise ValueError(f"Reference group `{ref_group}` is not present in the group labels.")
    unique_groups, group_counts = np.unique(groups, return_counts=True)
    groups_mapping = {g: i for i, g in enumerate(unique_groups)}
    encoded_groups = np.array([groups_mapping[g] for g in groups])
    # This should sort them in the same order as the first np.unique
    # TODO: np.lexsort so that indices are sorted within group as well. Just to be safer.^
    group_indices = np.argsort(groups)
    group_indptr = np.cumsum(np.insert(group_counts, 0, 0))

    return unique_groups, GroupContainer(
        encoded_groups=encoded_groups,
        counts=group_counts,
        indices=group_indices,
        indptr=group_indptr,
        encoded_ref_group=(
            -1 if ref_group is None else groups_mapping[ref_group]
        ),  # Weirdly enough, this must be -1 and not None, otherwise Numba fails to compile various functions, especially branching
    )



from typing import Literal

import numpy as np
from numba import njit

from illico.utils.groups import GroupContainer
from illico.utils.math import _add_at_scalar, compute_pval, diff
from illico.utils.ranking import _accumulate_group_ranksums_from_argsort
from illico.utils.sparse.csc import (
    CSCMatrix,
    _assert_is_csc,
    csc_fold_change,
    csc_get_cols,
)
from illico.utils.sparse.csr import (
    CSRMatrix,
    _assert_is_csr,
    csr_get_contig_cols_into_csc,
)


@njit(fastmath=True, nogil=True, cache=False)
def sparse_ovr_mwu_kernel(
    X: CSCMatrix,
    groups: np.ndarray,
    group_counts: np.ndarray,
    use_continuity: bool = True,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
) -> tuple[np.ndarray]:
    """Perform OVR ranksum test group wise and column wise on a CSC matrix.

    Args:
        X (CSCMatrix): CSC matrix holding expression counts.
        groups (np.ndarray): np.ndarray of shape (n_cells, ) holding encoded group labels.
        group_counts (np.ndarray): Count of cells per group.
        use_continuity (bool, optional): Apply continuity factor or not. Defaults to True.

    Returns:
        tuple[np.ndarray]: Two-sided p-values and U-statistics, per group and per gene (column).

    Author: Rémy Dubois
    """
    _, n_cols = X.shape
    # Convert n_zeros to float64 as they will be used for tie sum later
    n_zeros = (X.shape[0] - diff(X.indptr)).astype(np.float64)
    # Allocate placeholders for U stats and pvals
    U = np.empty((group_counts.size, n_cols), dtype=np.float64)
    pvals = np.empty((group_counts.size, n_cols), dtype=np.float64)

    # Note that because this function does not involve inner parallelism, this could be allocated per-col, but I find it cleaner this way
    nnz_per_group = np.zeros((group_counts.size, n_cols), dtype=np.float64)
    R1_nz = np.zeros((group_counts.size, n_cols), dtype=np.float64)
    # Note that if we run this function over chunks of columns, this work is repeated in each func but this is cheap
    n = group_counts.sum()
    n_ref = n - group_counts
    n_tgt = group_counts
    mu = n_ref * n_tgt / 2.0
    for j in range(n_cols):
        start, end = X.indptr[j], X.indptr[j + 1]
        nz_idx = X.indices[start:end]
        """Step 1: compute ranksum of non-zero elements, per group"""
        _idxs = np.argsort(X.data[start:end])
        tie_sum = _accumulate_group_ranksums_from_argsort(X.data[start:end], _idxs, groups[nz_idx], R1_nz[:, j])
        n0 = n_zeros[j]

        """Step 2: offset non-zero elements ranks by the number of zeros that precedes them"""
        if nz_idx.size:
            _add_at_scalar(nnz_per_group[:, j], groups[nz_idx], 1.0)
        # Deduce number of zeros per group
        nz_per_group = group_counts - nnz_per_group[:, j]
        # Offset the non-zero ranks by the amount of 0 that precedes them
        # All ranks must be shifted, so the sum is shifted by that many elements.
        R1_nz[:, j] += n0 * nnz_per_group[:, j]

        """ Step 3: Add ranksums of zero elements, per group"""
        # add zero contribution: number of zeros * avg rank
        R1 = R1_nz[:, j] + nz_per_group * (n0 + 1) / 2.0
        U[:, j] = n_ref * n_tgt + n_tgt * (n_tgt + 1) / 2 - R1
        tie_sum += n0**3 - n0

        for k in range(group_counts.size):
            pvals[k, j] = compute_pval(
                n_ref=n_ref[k],
                n_tgt=n_tgt[k],
                n=n,
                tie_sum=tie_sum,
                U=U[k, j],
                mu=mu[k],
                contin_corr=0.5 if use_continuity else 0.0,
                alternative=alternative,
            )

    return pvals, U


@njit(nogil=True, fastmath=True, cache=False)
def csc_ovr_mwu_kernel_over_contiguous_col_chunk(
    X: CSCMatrix,
    chunk_lb: int,
    chunk_ub: int,
    grpc: GroupContainer,
    is_log1p: bool,
    use_continuity: bool = True,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
):
    """Perform OVR ranksum test over the contiguous column chunk defined by the bounds.

    This function only applies to data stored in a CSC matrix.

    Args:
        X (CSCMatrix): Input CSC matrix holding expression counts
        chunk_lb (int): Lower bound of the vertial slice
        chunk_ub (int): Upper bound of the vertical slice
        grpc (GroupContainer): GroupContainer
        is_log1p (bool): User-indicated flag telling if data was log1p transformed or not.
        use_continuity (bool): Whether to use continuity correction when computing p-values.
        alternative (Literal["two-sided", "less", "greater"]): Type of alternative hypothesis.

    Raises:
        ValueError: If bounds are not intelligible

    Returns:
        tuple[np.ndarray]: two-sided p-values, u-statistics and fold changes,
        each of shape (n_groups, chunk_lb - chunk_ub).

    Author: Rémy Dubois
    """
    _assert_is_csc(X)
    if chunk_lb < 0 or chunk_ub > X.shape[1] or chunk_lb > chunk_ub:
        raise ValueError((chunk_lb, chunk_ub))

    csc_chunk = csc_get_cols(csc_matrix=X, indices=np.arange(chunk_lb, chunk_ub))

    # TODO: un-jitting this function comes at close to no cost, and allows to do argsorting out of the njit function
    # on linux machines, it is 3 to 4 times faster than numba.np.argsort and sorting seems to be half the compute time of the whole function
    # idxs = np.empty_like(csc_chunk.indices)
    # for j in range(csc_chunk.shape[1]):
    #     start, end = csc_chunk.indptr[j], csc_chunk.indptr[j + 1]
    #     idxs[start:end] = np.argsort(csc_chunk.data[start:end])
    pvalues, statistics = sparse_ovr_mwu_kernel(
        X=csc_chunk,
        groups=grpc.encoded_groups,
        group_counts=grpc.counts,
        use_continuity=use_continuity,
        alternative=alternative,
    )

    fold_change = csc_fold_change(X=csc_chunk, grpc=grpc, is_log1p=is_log1p)
    return pvalues, statistics, fold_change


@njit(nogil=True, fastmath=True, cache=False)  # This requires too many caching
def csr_ovr_mwu_kernel_over_contiguous_col_chunk(
    X: CSRMatrix,
    chunk_lb: int,
    chunk_ub: int,
    grpc: GroupContainer,
    is_log1p: bool,
    use_continuity: bool = True,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
) -> tuple[np.ndarray]:
    """Perform OVR ranksum test over the contiguous column chunk defined by the bounds.

    This function only applies to data stored in a CSR matrix.

    Args:
        X (CSRMatrix): Input CSR matrix holding expression counts
        chunk_lb (int): Lower bound of the vertial slice
        chunk_ub (int): Upper bound of the vertical slice
        grpc (GroupContainer): GroupContainer
        is_log1p (bool): User-indicated flag telling if data was log1p transformed or not.
        use_continuity (bool): Whether to use continuity correction when computing p-values.
        alternative (Literal["two-sided", "less", "greater"]): Type of alternative hypothesis

    Raises:
        ValueError: If bounds are not intelligible

    Returns:
        tuple[np.ndarray]: two-sided p-values, u-statistics and fold changes,
        each of shape (n_groups, chunk_lb - chunk_ub).

    Author: Rémy Dubois
    """
    _assert_is_csr(X)

    if chunk_lb < 0 or chunk_ub > X.shape[1] or chunk_lb > chunk_ub:
        raise ValueError((chunk_lb, chunk_ub))

    csc_chunk = csr_get_contig_cols_into_csc(csr_matrix=X, chunk_lb=chunk_lb, chunk_ub=chunk_ub)

    # TODO: same remark as csc regarding sorting
    pvalues, statistics = sparse_ovr_mwu_kernel(
        X=csc_chunk,
        groups=grpc.encoded_groups,
        group_counts=grpc.counts,
        use_continuity=use_continuity,
        alternative=alternative,
    )
    fold_change = csc_fold_change(X=csc_chunk, grpc=grpc, is_log1p=is_log1p)

    return pvalues, statistics, fold_change



import os
import time
from typing import Any

import numpy as np
import scipy.sparse as py_sparse
from loguru import logger
from numba import types

from illico.ovo.dense_ovo import dense_ovo_mwu_kernel_over_contiguous_col_chunk
from illico.ovo.sparse_ovo import (
    csc_ovo_mwu_kernel_over_contiguous_col_chunk,
    csr_ovo_mwu_kernel_over_contiguous_col_chunk,
)
from illico.ovr.dense_ovr import dense_ovr_mwu_kernel_over_contiguous_col_chunk
from illico.ovr.sparse_ovr import (
    csc_ovr_mwu_kernel_over_contiguous_col_chunk,
    csr_ovr_mwu_kernel_over_contiguous_col_chunk,
)
from illico.utils.groups import GroupContainer
from illico.utils.type import CSCMatrix, CSRMatrix

DISPATCHER_MAP = {
    ("dense", "ovr"): dense_ovr_mwu_kernel_over_contiguous_col_chunk,
    ("csc", "ovr"): csc_ovr_mwu_kernel_over_contiguous_col_chunk,
    ("csr", "ovr"): csr_ovr_mwu_kernel_over_contiguous_col_chunk,
    ("dense", "ovo"): dense_ovo_mwu_kernel_over_contiguous_col_chunk,
    ("csc", "ovo"): csc_ovo_mwu_kernel_over_contiguous_col_chunk,
    ("csr", "ovo"): csr_ovo_mwu_kernel_over_contiguous_col_chunk,
}


def _precompile(input_data: np.ndarray | py_sparse.csc_matrix | py_sparse.csr_matrix, reference_group: Any | None):
    """Precompile the CPU dispatcher before the threads start rushing to it.

    Note: a simpler way to do it could be to use a threading.lock to make just the first thread compile, but
    there are still concurrency risks.
    Having this in a separate routine also allows to call it from tests so that memory and speed benchmarks are not impacted by compilation.

    Args:
        input_data (np.ndarray | py_sparse.csc_matrix | py_sparse.csr_matrix): Input data
        reference_group (Any | None): Reference group

    Raises:
        ValueError: If input data is neither dense, CSC nor CSR.

    Author: Rémy Dubois
    """
    if os.environ.get("NUMBA_DISABLE_JIT", "0") == "1":
        logger.warning("Numba JIT is disabled, skipping precompilation.")
        return
    GroupContainerType = types.NamedTuple(
        [types.int64[::1], types.int64[::1], types.int64[::1], types.int64[::1], types.int64], GroupContainer
    )

    # This input signature corresponds to: lower bound, upper bvound, group container, is_log1p, use_continuity
    common_sig = (types.int64, types.int64, GroupContainerType, types.boolean, types.boolean, types.string)
    # This is the output: three float64 2D arrays
    out_sig = types.UniTuple(types.float64[:, ::1], 3)

    input_dtype = str(input_data.dtype)  # Sparse or dense expose .dtype
    if isinstance(input_data, np.ndarray):
        if input_data.flags["C_CONTIGUOUS"]:
            input_type = getattr(types, input_dtype)[:, ::1]
        else:
            input_type = getattr(types, input_dtype)[:, :]
        data_format = "dense"
    elif isinstance(input_data, py_sparse.spmatrix):
        # TODO: are inputs necessarily C-contiguous?
        input_type = getattr(types, input_dtype)[::1]
        indices_type = getattr(types, str(input_data.indices.dtype))[::1]
        indptr_type = getattr(types, str(input_data.indptr.dtype))[::1]
        if isinstance(input_data, py_sparse.csc_matrix):
            input_type = types.NamedTuple(
                [input_type, indices_type, indptr_type, types.UniTuple(types.int64, 2)], CSCMatrix
            )
            data_format = "csc"
        elif isinstance(input_data, py_sparse.csr_matrix):
            input_type = types.NamedTuple(
                [input_type, indices_type, indptr_type, types.UniTuple(types.int64, 2)], CSRMatrix
            )
            data_format = "csr"
        else:
            raise ValueError(f"Unsupported sparse matrix type {type(input_data)}")
    else:
        raise ValueError(f"Unsupported input data type {type(input_data)}")

    if reference_group is None:
        test_type = "ovr"
    else:
        test_type = "ovo"

    dispatcher = DISPATCHER_MAP[(data_format, test_type)]
    sig = out_sig(input_type, *common_sig)

    s = time.time()
    dispatcher.compile(sig)
    e = time.time()
    logger.trace(f"Precompilation of {data_format}-{test_type} dispatcher took {e - s:.1f}s")
    return dispatcher



from collections import namedtuple

import numpy as np
from scipy import sparse

TestResults = namedtuple("TestResults", ["statistic", "pvalue"])
CSRMatrix = namedtuple("CSRMatrix", ["data", "indices", "indptr", "shape"])
CSCMatrix = namedtuple("CSCMatrix", ["data", "indices", "indptr", "shape"])


def scipy_to_nb(
    obj: np.ndarray | sparse.csc_matrix | sparse.csr_matrix,
) -> np.ndarray | CSRMatrix | CSCMatrix:
    """Convert python object into Numba-compat holder

    Returns
    -------
    np.ndarray | CSRMatrix | CSCMatrix
        The Numba-compatible object

    Raises
    ------
    ValueError
        If input is neither dense, csc or csr.

    Author: Rémy Dubois
    """
    if isinstance(obj, np.ndarray):
        return obj
    elif isinstance(obj, sparse.csc_matrix):
        return CSCMatrix(obj.data, obj.indices, obj.indptr, obj.shape)
    elif isinstance(obj, sparse.csr_matrix):
        return CSRMatrix(obj.data, obj.indices, obj.indptr, obj.shape)
    else:
        raise ValueError(type(obj))



from typing import Literal

import numpy as np
from numba import njit

from illico.utils.groups import GroupContainer
from illico.utils.math import compute_pval, diff
from illico.utils.ranking import (
    _sort_csc_columns_inplace,
    rank_sum_and_ties_from_sorted,
)
from illico.utils.sparse.csc import csc_get_contig_cols_into_csr
from illico.utils.sparse.csr import (
    csr_fold_change,
    csr_get_contig_cols_into_csr,
    csr_get_rows_into_csc,
)
from illico.utils.type import CSCMatrix, CSRMatrix


@njit(nogil=True, fastmath=True, cache=False)
def single_group_sparse_ovo_mwu_kernel(
    sorted_ref_data: CSCMatrix,
    sorted_tgt_data: CSCMatrix,
    use_continuity: bool = True,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
) -> tuple[np.ndarray]:
    """Perform OVO tests gene wise using the two **sorted** CSC matrix given as input.

    The test performed is the equivalent of:
    `scipy.stats.mannwhitneyu(sorted_ref_data.toarray(), sorted_tgt_data.toarray(), use_continuity=True)`

    Args:
        sorted_ref_data (CSCMatrix): Reference data stored in CSC, sorted column-wise
        sorted_tgt_data (CSCMatrix): Perturbed data stored in CSC, sorted column-wise
        use_continuity (bool, optional): Apply continuity factor or not. Defaults to True.
        alternative (Literal["two-sided", "less", "greater"]): Type of alternative hypothesis

    Raises:
        ValueError: If shape mismatche

    Returns:
        tuple[np.ndarray]: two-sided p-values, U-statistics. Each of shape (n_genes,).

    Author: Rémy Dubois
    """

    n_ref, n_cols_ref = sorted_ref_data.shape
    n_tgt, n_cols_tgt = sorted_tgt_data.shape
    # TODO: make this check everywhere or nowhere
    if n_cols_ref != n_cols_tgt:
        raise ValueError("Uneven number of columns between ref and perturbed.")

    # Allocate placeholders
    n_zeros_tgt = (n_tgt - diff(sorted_tgt_data.indptr)).astype(np.int64)
    n_zeros_ref = (n_ref - diff(sorted_ref_data.indptr)).astype(np.int64)
    U_statistics = np.empty(n_cols_ref, dtype=np.float64)
    pvals = np.empty(n_cols_ref, dtype=np.float64)
    n = n_ref + n_tgt
    mu = n_ref * n_tgt / 2.0
    for j in range(n_cols_ref):
        n_zeros_combined = n_zeros_ref[j] + n_zeros_tgt[j]
        # Get the bounds
        lbt, ubt = sorted_tgt_data.indptr[j], sorted_tgt_data.indptr[j + 1]
        lbr, ubr = sorted_ref_data.indptr[j], sorted_ref_data.indptr[j + 1]

        # Compute ranksum and tie sum for non zero values
        ranksum, tie_sum = rank_sum_and_ties_from_sorted(sorted_ref_data.data[lbr:ubr], sorted_tgt_data.data[lbt:ubt])

        # Offset the ranks of the number of zeros in ref and perturbed
        ranksum += n_zeros_combined * (ubt - lbt)

        # Compute ranksum
        n0 = n_zeros_tgt[j]
        R1_nz = ranksum  # Sum ranks
        R1 = R1_nz + n0 * (n_zeros_ref[j] + n0 + 1) / 2.0  # Add sumranks of zeros

        # Compute U-stat
        U1 = n_ref * n_tgt + n_tgt * (n_tgt + 1) / 2 - R1

        # Compute sigma
        tie_sum += n_zeros_combined**3 - n_zeros_combined
        pvals[j] = compute_pval(
            n_ref=n_ref,
            n_tgt=n_tgt,
            n=n,
            tie_sum=tie_sum,
            U=U1,
            mu=mu,
            contin_corr=0.5 if use_continuity else 0.0,
            alternative=alternative,
        )

        # Regardless of the alternative, always return U1 like scipy
        U_statistics[j] = U1

    return pvals, U_statistics


@njit(nogil=True, fastmath=True, cache=False)
def multi_group_sparse_ovo_mwu_kernel(
    X: CSRMatrix,
    grpc: GroupContainer,
    ref_group_id: int,
    use_continuity: bool = True,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
) -> tuple[np.ndarray]:
    """Sequentially perform group-wise OVO tests along the columns on a CSR matrix.

    Args:
        X (CSRMatrix): Input CSR matrix of shape (n_cells, n_genes)
        grpc (GroupContainer): GroupContainer
        ref_group_id (int): Encoded reference group ID.
        use_continuity (bool, optional): Whether to use continuity correction when computing p-values. Defaults to True.
        alternative (Literal["two-sided", "less", "greater"]): Type of alternative hypothesis

    Returns:
        tuple[np.ndarray]: two-sided p-values, U-statistics. Each of shape (n_groups, n_genes).

    Author: Rémy Dubois
    """
    # Avoid accessing attributes in the for loop
    group_indices = grpc.indices
    group_indptr = grpc.indptr
    # Now, get the ref group, and convert it back to CSC because the test requires CSC data
    n_groups = group_indptr.size - 1
    ref_indices = group_indices[group_indptr[ref_group_id] : group_indptr[ref_group_id + 1]]
    csc_X_ref = csr_get_rows_into_csc(X, ref_indices)
    _sort_csc_columns_inplace(csc_matrix=csc_X_ref)

    # Now go through all the groups one by one
    pvalues = np.empty((n_groups, X.shape[1]), dtype=np.float64)
    statistics = np.empty((n_groups, X.shape[1]), dtype=np.float64)
    for group_id in range(group_indptr.size - 1):
        if group_id == ref_group_id:
            pvalues[group_id, :] = 1.0
            statistics[group_id, :] = -1.0
            continue

        tgt_idxs = group_indices[group_indptr[group_id] : group_indptr[group_id + 1]]
        X_tgt = csr_get_rows_into_csc(X, tgt_idxs)
        _sort_csc_columns_inplace(X_tgt)
        pvalue, statistic = single_group_sparse_ovo_mwu_kernel(
            sorted_ref_data=csc_X_ref, sorted_tgt_data=X_tgt, use_continuity=use_continuity, alternative=alternative
        )
        pvalues[group_id, :] = pvalue
        statistics[group_id, :] = statistic

    return pvalues, statistics


# Not jitting this and sorting all the cells at once is 1.5x slower. Ideally, it would be faster to sort only groups one by one but
# doubt this would be enough faster (think of mergesort) => It is twice faster, so i dont think it will bridge the gap
@njit(nogil=True, fastmath=True, cache=False)  # This requires too many caching, too dangerous
def csc_ovo_mwu_kernel_over_contiguous_col_chunk(
    X: CSCMatrix,
    chunk_lb: int,
    chunk_ub: int,
    grpc: GroupContainer,
    is_log1p: bool,
    use_continuity: bool = True,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
):
    """Perform OVO test over contiguous chunk of column of a CSC matrix.

    Args:
        X (CSCMatrix): Input matrix of shape (n_cells, n_genes)
        chunk_lb (int): Lower bound of the vertical slicing
        chunk_ub (int): Upper bound of the vertical slicing
        grpc (GroupContainer): GroupContainer
        is_log1p (bool): User-indicated flag telling if data underwent log1p transform.

    Raises:
        ValueError: If chunk bounds are inintelligible.

    Returns:
        tuple[np.ndarray]: two-sided p-values, U-statistics, fold change. Each
        of shape (n_groups, chunk_ub - chunk_lb).

    Author: Rémy Dubois
    """
    if chunk_lb < 0 or chunk_ub > X.shape[1] or chunk_lb > chunk_ub:
        raise ValueError((chunk_lb, chunk_ub))

    # This copies the data, so all that follow can happen in-place
    csr_chunk = csc_get_contig_cols_into_csr(X, chunk_lb, chunk_ub)
    pvalues, statistics = multi_group_sparse_ovo_mwu_kernel(
        X=csr_chunk,
        grpc=grpc,
        ref_group_id=grpc.encoded_ref_group,
        use_continuity=use_continuity,
        alternative=alternative,
    )

    # Compute fold change
    fold_change = csr_fold_change(csr_chunk, grpc, is_log1p=is_log1p)

    return pvalues, statistics, fold_change


# Real scale tests on whole H1 showed 24secs on 8 threads and 2min45s on 1, so a speedup of 165 / 24 = 6.875x
@njit(nogil=True, fastmath=True, cache=False)
def csr_ovo_mwu_kernel_over_contiguous_col_chunk(
    X: CSRMatrix,
    chunk_lb: int,
    chunk_ub: int,
    grpc: GroupContainer,
    is_log1p: bool,
    use_continuity: bool = True,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
):
    """Perform OVO test over contiguous chunk of column of a CSR matrix.

    Args:
        X (CSRMatrix): Input matrix of shape (n_cells, n_genes)
        chunk_lb (int): Lower bound of the vertical slicing
        chunk_ub (int): Upper bound of the vertical slicing
        grpc (GroupContainer): GroupContainer
        is_log1p (bool): User-indicated flag telling if data underwent log1p transform.
        use_continuity (bool): Whether to use continuity correction when computing p-values.
        alternative (Literal["two-sided", "less", "greater"]): Type of alternative hypothesis

    Raises:
        ValueError: If chunk bounds are inintelligible.

    Returns:
        tuple[np.ndarray]: two-sided p-values, U-statistics, fold change. Each
        of shape (n_groups, chunk_ub - chunk_lb).

    Author: Rémy Dubois
    """
    if chunk_lb < 0 or chunk_ub > X.shape[1] or chunk_lb > chunk_ub:
        raise ValueError((chunk_lb, chunk_ub))

    # This copies the data, so all that follow can happen in-place
    csr_chunk = csr_get_contig_cols_into_csr(X, chunk_lb, chunk_ub)
    pvalues, statistics = multi_group_sparse_ovo_mwu_kernel(
        X=csr_chunk,
        grpc=grpc,
        ref_group_id=grpc.encoded_ref_group,
        use_continuity=use_continuity,
        alternative=alternative,
    )
    # Compute fold change
    fold_change = csr_fold_change(csr_chunk, grpc, is_log1p=is_log1p)

    return pvalues, statistics, fold_change



import numpy as np
from numba import njit

from illico.utils.groups import GroupContainer
from illico.utils.math import _add_at_scalar, _add_at_vec, fold_change_from_summed_expr
from illico.utils.type import CSCMatrix, CSRMatrix


@njit(nogil=True, cache=False)
def _assert_is_csc(matrix: CSCMatrix | CSRMatrix) -> None:
    """Assert a matrix is CSC.

    Args:
        obj (CSCMatrix|CSRMatrix): The matrix to check.

    Author: Rémy Dubois
    """
    n_parcels = matrix.indptr.size - 1
    _, n_cols = matrix.shape
    assert n_parcels == n_cols, "Matric is not CSC: indptr and number of columns don't match."


@njit(nogil=True, fastmath=True, cache=False)
def csc_count_nonzeros(csc_matrix: CSCMatrix, axis: int | None = None) -> np.ndarray:
    """Cout number of non zero values in a CSC matrix.

    Args:
        csc_matrix (CSCMatrix): Input matrix
        axis (int | None, optional): Axis along which to compute. Defaults to None.

    Raises:
        ValueError: If axis is not intelligible.

    Returns:
        np.ndarray: Number of non-zero values, always a 1-d array for compilation purposes.

    Author: Rémy Dubois
    """
    if axis is None:
        nnz = np.empty((1,), dtype=np.int32)
        nnz[0] = csc_matrix.data.size
        return nnz
    elif axis == 1:
        nnz = np.zeros(csc_matrix.shape[0], dtype=np.int32)
        _add_at_scalar(nnz, csc_matrix.indices, 1)
        return nnz
    elif axis == 0:
        # Idk why, np.diff crashes
        nnz = np.zeros(csc_matrix.shape[1], dtype=np.int32)
        for i in range(1, csc_matrix.indptr.size):
            nnz[i - 1] = csc_matrix.indptr[i] - csc_matrix.indptr[i - 1]
        return nnz
    else:
        raise ValueError(axis)


@njit(nogil=True, fastmath=True, cache=False)
def csc_to_csr(csc_matrix: CSCMatrix) -> CSRMatrix:
    """Convert a CSC matrix to CSR.

    Args:
        csc_matrix (CSCMatrix): Input CSC matrix

    Returns:
        CSRMatrix: The resulting CSR matrix.

    Author: Rémy Dubois
    """
    nnz = csc_matrix.data.size
    csr_indptr = np.zeros(csc_matrix.shape[0] + 1, dtype=csc_matrix.indptr.dtype)
    csr_indices = np.empty(nnz, dtype=csc_matrix.indices.dtype)
    csr_data = np.empty(nnz, dtype=csc_matrix.data.dtype)

    # Pass 1: count number of entries per row
    csr_indptr[1:] = np.cumsum(csc_count_nonzeros(csc_matrix, 1))
    next_row = csr_indptr.copy()

    # Pass 2: fill CSC structure
    for col in range(csc_matrix.shape[1]):
        col_start = csc_matrix.indptr[col]
        col_end = csc_matrix.indptr[col + 1]
        for idx in range(col_start, col_end):
            row = csc_matrix.indices[idx]
            dest = next_row[row]
            csr_indices[dest] = col
            csr_data[dest] = csc_matrix.data[idx]
            next_row[row] += 1
    return CSRMatrix(csr_data, csr_indices, csr_indptr, csc_matrix.shape)


@njit(nogil=True, fastmath=True, cache=False)
def csc_get_cols(csc_matrix: CSCMatrix, indices: np.ndarray) -> CSCMatrix:
    """Perform vertical slicing of a CSC matrix.

    Equivalent of `scipy.sparse.csc_matrix[:, indices]`.

    Args:
        csc_matrix (CSCMatrix): Input CSC matrix
        indices (np.ndarray): Indices of the columns to get.

    Raises:
        ValueError: If indices are not intelligible

    Returns:
        CSCMatrix: The resulting (sliced) CSC matrix

    Author: Rémy Dubois
    """
    if indices.min() < 0 or indices.max() > csc_matrix.shape[0]:
        raise ValueError(indices.min(), indices.max())
    new_indptr = np.empty(indices.size + 1, dtype=csc_matrix.indptr.dtype)
    new_indptr[0] = 0
    for i in range(indices.size):
        col_idx = indices[i]
        col_size = csc_matrix.indptr[col_idx + 1] - csc_matrix.indptr[col_idx]
        new_indptr[i + 1] = col_size + new_indptr[i]

    new_data = np.empty(new_indptr[-1], csc_matrix.data.dtype)
    new_indices = np.empty(new_indptr[-1], csc_matrix.indices.dtype)
    for i in range(indices.size):
        col_idx = indices[i]
        new_data[new_indptr[i] : new_indptr[i + 1]] = csc_matrix.data[
            csc_matrix.indptr[col_idx] : csc_matrix.indptr[col_idx + 1]
        ]
        new_indices[new_indptr[i] : new_indptr[i + 1]] = csc_matrix.indices[
            csc_matrix.indptr[col_idx] : csc_matrix.indptr[col_idx + 1]
        ]
    return CSCMatrix(new_data, new_indices, new_indptr, (csc_matrix.shape[0], indices.size))


@njit(nogil=True, fastmath=True, cache=False)
def csc_get_contig_cols_into_csr(csc_matrix: CSCMatrix, chunk_lb: int, chunk_ub: int) -> CSRMatrix:
    """Perform contiguous vertical slicing of a CSC matrix and stores it into a CSR matrix.

    This function is the equivalent of `scipy.sparse.csc_matrix[:, chunk_lb:chunk_ub].to_csr()`.

    Args:
        csc_matrix (CSCMatrix): Input CSC matrix
        chunk_lb (int): Lower bound of the vertical slicing
        chunk_ub (int): Upper bound of the vertical slicing

    Raises:
        ValueError: If bounds are not intelligible.

    Returns:
        CSRMatrix: Resulting CSR matrix

    Author: Rémy Dubois
    """
    if chunk_lb < 0 or chunk_ub > csc_matrix.shape[1] or chunk_lb > chunk_ub:
        raise ValueError((chunk_lb, chunk_ub))

    csr_nnz = np.zeros(csc_matrix.shape[0], dtype=np.int64)
    _add_at_scalar(
        csr_nnz,
        csc_matrix.indices[csc_matrix.indptr[chunk_lb] : csc_matrix.indptr[chunk_ub]],
        1,
    )
    # Of shape nrows+1, with a leading 0
    csr_indptr = np.cumsum(np.roll(np.append(csr_nnz, 0), 1))

    # Allocate placeholders for data and indices
    new_data = np.empty(csr_indptr[-1], csc_matrix.data.dtype)
    new_indices = np.empty(csr_indptr[-1], csc_matrix.indices.dtype)
    pointer = csr_indptr.copy()  # Keep one pointer per row, indicating where to store the next value
    for j in range(chunk_lb, chunk_ub):
        col_start = csc_matrix.indptr[j]
        col_end = csc_matrix.indptr[j + 1]
        for idx in range(col_start, col_end):
            row_idx = csc_matrix.indices[idx]
            v = csc_matrix.data[idx]
            new_data[pointer[row_idx]] = v
            new_indices[pointer[row_idx]] = j - chunk_lb
            pointer[row_idx] += 1
    return CSRMatrix(new_data, new_indices, csr_indptr, (csc_matrix.shape[0], chunk_ub - chunk_lb))


@njit(nogil=True, fastmath=True, cache=False)
def csc_fold_change(X: CSCMatrix, grpc: GroupContainer, is_log1p: bool) -> np.ndarray:
    """Compute fold change from a CSC matrix of expression counts.

    Args:
        X (CSCMatrix): Input expression counts CSC matrix
        grpc (GroupContainer): GroupContainer
        is_log1p (bool): User-indicated flag telling if data was log1p or not.

    Returns:
        np.ndarray: Fold change of change (n_groups, n_genes)

    Author: Rémy Dubois
    """
    _assert_is_csc(X)
    group_agg_counts = np.zeros(shape=(grpc.counts.size, X.shape[1]), dtype=np.float64)
    # Sum expressions per group
    for j in range(X.shape[1]):
        start = X.indptr[j]
        end = X.indptr[j + 1]
        row_indices = X.indices[start:end]
        row_data = np.expm1(X.data[start:end]) if is_log1p else X.data[start:end]
        group_id = grpc.encoded_groups[row_indices]
        _add_at_vec(group_agg_counts[:, j], group_id, row_data)
    fold_change = fold_change_from_summed_expr(group_agg_counts, grpc)
    return fold_change



import numpy as np
from numba import njit

from illico.utils.groups import GroupContainer
from illico.utils.math import (
    _add_at_scalar,
    _add_at_vec,
    diff,
    fold_change_from_summed_expr,
)
from illico.utils.type import CSCMatrix, CSRMatrix


@njit(nogil=True, cache=False)
def _assert_is_csr(matrix: CSCMatrix | CSRMatrix) -> None:
    """Assert a matrix is CSR.

    Args:
        matrix (CSCMatrix | CSRMatrix): Input matrix to check.

    Author: Rémy Dubois
    """
    n_parcels = matrix.indptr.size - 1
    n_rows, _ = matrix.shape
    assert n_parcels == n_rows, "Object is not CSR: indptr and number of rows don't match."


@njit(nogil=True, fastmath=True, cache=False)
def csr_count_nonzeros(csr_matrix: CSRMatrix, axis: int | None = None) -> np.ndarray:
    """Count number of non zeros in a CSR matrix.

    Args:
        csr_matrix (CSRMatrix): Input matrix
        axis (int | None, optional): Axis along which to compute. Defaults to None.

    Raises:
        ValueError: If axis is not intelligible.

    Returns:
        np.ndarray: Number of non-zero values, always a 1-d array for compilation purposes.

    Author: Rémy Dubois
    """
    if axis is None:
        nnz = np.empty((1,), dtype=np.int32)
        nnz[0] = csr_matrix.data.size
        return nnz
    elif axis == 0:
        nnz = np.zeros(csr_matrix.shape[1], dtype=np.int32)
        _add_at_scalar(nnz, csr_matrix.indices, 1)
        return nnz
    elif axis == 1:
        nnz = diff(csr_matrix.indptr)
        return nnz
    else:
        raise ValueError(axis)


@njit(nogil=True, fastmath=True, cache=False)
def csr_to_csc(csr_matrix: CSRMatrix) -> CSCMatrix:
    """Convert a CSR matrix to CSC.

    Args:
        csr_matrix (CSRMatrix): Input CSR matrix

    Returns:
        CSCMatrix: The resulting CSC matrix.

    Author: Rémy Dubois
    """
    nnz = csr_matrix.data.size

    # Allocate placeholders
    csc_indptr = np.zeros(csr_matrix.shape[1] + 1, dtype=csr_matrix.indptr.dtype)
    csc_indices = np.empty(nnz, dtype=csr_matrix.indices.dtype)
    csc_data = np.empty(nnz, dtype=csr_matrix.data.dtype)

    # Pass 1: count number of entries per column
    csc_indptr[1:] = np.cumsum(csr_count_nonzeros(csr_matrix, 0))
    next_col = csc_indptr.copy()

    # Pass 2: fill CSC structure
    for row in range(csr_matrix.shape[0]):
        row_start = csr_matrix.indptr[row]
        row_end = csr_matrix.indptr[row + 1]
        for idx in range(row_start, row_end):
            col = csr_matrix.indices[idx]
            dest = next_col[col]
            csc_indices[dest] = row
            csc_data[dest] = csr_matrix.data[idx]
            next_col[col] += 1
    return CSCMatrix(csc_data, csc_indices, csc_indptr, csr_matrix.shape)


@njit(nogil=True, fastmath=True, cache=False)
def csr_get_rows_into_csc(csr_matrix: CSRMatrix, indices: np.ndarray) -> CSCMatrix:
    """Performs horizontal slicing of a CSR matrix and stores it in a CSC matrix.

    Equivalent of `scipy.sparse.csr_matrix[indices, :].to_csc()`.

    Args:
        csr_matrix (CSRMatrix): Input CSR matrix
        indices (np.ndarray): Indices of the rows to get.

    Returns:
        CSCMatrix: The resulting CSC matrix holding data of the sliced rows.

    Author: Rémy Dubois
    """
    # Count non zeros per col
    csc_nnz = np.zeros(csr_matrix.shape[1], dtype=np.int64)
    for row_idx in indices:
        _add_at_scalar(
            csc_nnz,
            csr_matrix.indices[csr_matrix.indptr[row_idx] : csr_matrix.indptr[row_idx + 1]],
            1,
        )
    # Of shape ncols +1, with leading 0
    csc_indptr = np.cumsum(np.roll(np.append(csc_nnz, 0), 1))
    new_data = np.empty(csc_indptr[-1], csr_matrix.data.dtype)
    new_indices = np.empty(csc_indptr[-1], csr_matrix.indices.dtype)
    pointer = csc_indptr.copy()  # Keep one pointer per row, indicating where to store the next value
    for i in range(indices.size):
        row_idx = indices[i]
        row_start = csr_matrix.indptr[row_idx]
        row_end = csr_matrix.indptr[row_idx + 1]
        for idx in range(row_start, row_end):
            col_idx = csr_matrix.indices[idx]
            v = csr_matrix.data[idx]
            new_data[pointer[col_idx]] = v
            new_indices[pointer[col_idx]] = i
            pointer[col_idx] += 1
    return CSCMatrix(new_data, new_indices, csc_indptr, (indices.size, csr_matrix.shape[1]))


@njit(nogil=True, fastmath=True, cache=False)
def csr_get_contig_cols_into_csr(csr_matrix: CSRMatrix, lb: int, ub: int) -> CSRMatrix:
    """Perform contiguous vertical slicing of a CSR matrix and stores it into a CSR matrix.

    Args:
        csr_matrix (CSRMatrix): Input CSR matrix.
        lb (int): Lower bound of the vertical slicing.
        ub (int): Upper bound of the vertical slicing.

    Raises:
        ValueError: If bounds are not intelligible.

    Returns:
        CSRMatrix: The CSR matrix holding data of the sliced columns.

    Author: Rémy Dubois
    """
    if lb < 0 or ub > csr_matrix.shape[1] or lb > ub:
        raise ValueError((lb, ub))

    # Start by counting non zeros, and store each column's parcel in the csr arrays in the bounds var
    n_nzeros = np.zeros(csr_matrix.shape[0], dtype=np.int32)
    bounds = np.empty((csr_matrix.shape[0], 2), dtype=np.int32)
    for i in range(csr_matrix.shape[0]):
        start, end = csr_matrix.indptr[i], csr_matrix.indptr[i + 1]
        # Log instead of linear search, this is the only real optimization compared to `sparse.csc_matrix(csr_mat[:, slice])`
        # did not really pay attention to `side` but it seems to work as it is now.
        cb, rb = np.searchsorted(csr_matrix.indices[start:end], [lb, ub])
        bounds[i, 0] = cb
        bounds[i, 1] = rb
        n_nzeros[i] += rb - cb

    # # TODO: unify this with indptr
    chunk_offset_per_col = np.roll(n_nzeros, 1)
    chunk_offset_per_col[0] = 0
    chunk_offset_per_col = np.cumsum(chunk_offset_per_col)

    # # Local counter of how many elements where stored in result's columns
    data_placeholder = np.empty(n_nzeros.sum(), dtype=csr_matrix.data.dtype)
    indices_placeholder = np.empty(n_nzeros.sum(), dtype=csr_matrix.indices.dtype)
    counter = 0
    # Go through all the nonzero elements
    for i in range(csr_matrix.shape[0]):
        start, end = csr_matrix.indptr[i], csr_matrix.indptr[i + 1]
        cb, rb = bounds[i, 0], bounds[i, 1]
        chunk_data = csr_matrix.data[start + cb : start + rb]
        data_placeholder[counter : counter + chunk_data.size] = chunk_data.copy()
        chunk_indices = csr_matrix.indices[start + cb : start + rb] - lb
        indices_placeholder[counter : counter + chunk_indices.size] = chunk_indices.copy()
        counter += rb - cb

    indptr = np.append(chunk_offset_per_col, data_placeholder.size).astype(csr_matrix.indptr.dtype)
    return CSRMatrix(data_placeholder, indices_placeholder, indptr, (csr_matrix.shape[0], ub - lb))


@njit(nogil=True, fastmath=True, cache=False)
def csr_get_contig_cols_into_csc(csr_matrix: CSRMatrix, chunk_lb: int, chunk_ub: int) -> CSCMatrix:
    """Perform contiguous vertical slicing of a CSR matrix and stores it into a CSC matrix.

    Args:
        csr_matrix (CSRMatrix): Input CSR matrix.
        lb (int): Lower bound of the vertical slicing.
        ub (int): Upper bound of the vertical slicing.

    Raises:
        ValueError: If bounds are not intelligible.

    Returns:
        CSCMatrix: The CSC matrix holding data of the sliced columns.

    Author: Rémy Dubois
    """

    if chunk_lb < 0 or chunk_ub > csr_matrix.shape[1] or chunk_lb > chunk_ub:
        raise ValueError((chunk_lb, chunk_ub))
    # Start by counting non zeros, and store each column's parcel in the csr arrays in the bounds var
    n_nzeros = np.zeros(chunk_ub - chunk_lb, dtype=csr_matrix.indptr.dtype)
    bounds = np.empty((csr_matrix.shape[0], 2), dtype=np.int64)
    for i in range(csr_matrix.shape[0]):
        start, end = csr_matrix.indptr[i], csr_matrix.indptr[i + 1]
        # Log instead of linear search, this is the only real optimization compared to `sparse.csc_matrix(csr_mat[:, slice])`
        # did not really pay attention to `side` but it seems to work as it is now.
        cb, rb = np.searchsorted(csr_matrix.indices[start:end], [chunk_lb, chunk_ub])
        bounds[i, 0] = cb
        bounds[i, 1] = rb
        for j in range(start + cb, start + rb):
            n_nzeros[csr_matrix.indices[j] - chunk_lb] += 1

    # # TODO: unify this with indptr
    chunk_offset_per_col = np.roll(n_nzeros, 1)
    chunk_offset_per_col[0] = 0
    chunk_offset_per_col = np.cumsum(chunk_offset_per_col)
    # Local counter of how many elements where stored in result's columns
    data_placeholder = np.empty(n_nzeros.sum(), dtype=csr_matrix.data.dtype)
    indices_placeholder = np.empty(n_nzeros.sum(), dtype=csr_matrix.indices.dtype)
    col_counter = np.zeros(csr_matrix.shape[1], dtype=np.int64)
    # Go through all the nonzero elements
    for i in range(csr_matrix.shape[0]):
        start, end = csr_matrix.indptr[i], csr_matrix.indptr[i + 1]
        cb, rb = bounds[i, 0], bounds[i, 1]
        for j in range(start + cb, start + rb):
            col_idx = csr_matrix.indices[j] - chunk_lb
            ph_idx = col_counter[col_idx] + chunk_offset_per_col[col_idx]
            data_placeholder[ph_idx] = csr_matrix.data[j]
            indices_placeholder[ph_idx] = i
            col_counter[col_idx] += 1

    indptr = np.append(chunk_offset_per_col, data_placeholder.size).astype(csr_matrix.indptr.dtype)
    return CSCMatrix(
        data_placeholder,
        indices_placeholder,
        indptr,
        (csr_matrix.shape[0], chunk_ub - chunk_lb),
    )


# TODO: move this in the same file as its subroutines, so that caching as no risk of staling
@njit(nogil=True, fastmath=True, cache=False)
def csr_fold_change(X: CSRMatrix, grpc: GroupContainer, is_log1p: bool) -> np.ndarray:
    """Compute fold change from a CSR matrix of expression counts.

    Args:
        X (CSRMatrix): Input expression counts CSR matrix
        grpc (GroupContainer): GroupContainer
        is_log1p (bool): User-indicated flag telling if data was log1p or not.

    Returns:
        np.ndarray: Fold change of change (n_groups, n_genes)

    Author: Rémy Dubois
    """
    _assert_is_csr(X)
    group_agg_counts = np.zeros(shape=(grpc.counts.size, X.shape[1]), dtype=np.float64)
    # Sum expressions per group
    for i in range(X.shape[0]):
        start = X.indptr[i]
        end = X.indptr[i + 1]
        col_indices = X.indices[start:end]
        row_data = np.expm1(X.data[start:end]) if is_log1p else X.data[start:end]
        group_id = grpc.encoded_groups[i]
        _add_at_vec(group_agg_counts[group_id], col_indices, row_data)
    fold_change = fold_change_from_summed_expr(group_agg_counts, grpc)
    return fold_change



from typing import Literal

import numpy as np
from scipy import sparse as py_sparse

from illico.ovo.dense_ovo import dense_ovo_mwu_kernel_over_contiguous_col_chunk
from illico.ovo.sparse_ovo import (
    csc_ovo_mwu_kernel_over_contiguous_col_chunk,
    csr_ovo_mwu_kernel_over_contiguous_col_chunk,
)
from illico.utils.groups import GroupContainer
from illico.utils.sparse.csc import CSCMatrix
from illico.utils.sparse.csr import CSRMatrix
from illico.utils.type import scipy_to_nb


def ovo_mwu_over_contiguous_col_chunk(
    X: np.ndarray | py_sparse.csc_matrix | py_sparse.csr_matrix,
    chunk_lb: int,
    chunk_ub: int,
    group_container: GroupContainer,
    is_log1p: bool,
    use_continuity: bool = True,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
):
    """Dispatcher for the OVO ranksum test. Routes toward the optimized
    implementation depending on input data format.

    Args:
        X (np.ndarray | py_sparse.csc_matrix | py_sparse.csr_matrix): Input expression count matrix.
        chunk_lb (int): Vertical slicing (column) lower bound
        chunk_ub (int): Vertical slicing (column) upper bound
        group_container (GroupContainer): GroupContainer
        is_log1p (bool): User-indicated flag telling if expression counts underwent log1p transfo.
        use_continuity (bool, optional): Whether to use continuity correction when computing p-values.
        alternative (Literal["two-sided", "less", "greater"]): Type of alternative hypothesis.

    Raises:
        ValueError: If input is neither dense, csc or csr.

    Returns:
        tuple[np.ndarray]: Two-sided p-values, U-statistics and fold change.
        Each np.ndarray of shape (n_groups, n_genes)

    Author: Rémy Dubois
    """
    # Convert to numba friendly dtype
    X_nb = scipy_to_nb(X)

    # Check all possible input types
    if isinstance(X, np.ndarray):
        pvalues, statistics, fold_change = dense_ovo_mwu_kernel_over_contiguous_col_chunk(
            X,
            chunk_lb,
            chunk_ub,
            group_container,
            is_log1p=is_log1p,
            use_continuity=use_continuity,
            alternative=alternative,
        )
    elif isinstance(X_nb, CSCMatrix):
        pvalues, statistics, fold_change = csc_ovo_mwu_kernel_over_contiguous_col_chunk(
            X_nb,
            chunk_lb,
            chunk_ub,
            group_container,
            is_log1p=is_log1p,
            use_continuity=use_continuity,
            alternative=alternative,
        )
    elif isinstance(X_nb, CSRMatrix):
        pvalues, statistics, fold_change = csr_ovo_mwu_kernel_over_contiguous_col_chunk(
            X_nb,
            chunk_lb,
            chunk_ub,
            group_container,
            is_log1p=is_log1p,
            use_continuity=use_continuity,
            alternative=alternative,
        )
    else:
        raise ValueError(X_nb)

    return pvalues, statistics, fold_change



from typing import Literal

import numpy as np
from numba import njit

from illico.utils.groups import GroupContainer
from illico.utils.math import chunk_and_fortranize, compute_pval, dense_fold_change
from illico.utils.ranking import (
    _sort_along_axis_inplace,
    rank_sum_and_ties_from_sorted,
)


@njit(nogil=True, fastmath=True, parallel=False, cache=False)
def dense_ovo_mwu_kernel(
    sorted_ref_data: np.ndarray,
    sorted_tgt_data: np.ndarray,
    use_continuity: bool = True,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
) -> tuple[np.ndarray]:
    """Sequentially perform OVO tests on columns between sorted ref and sorted perturbed data.

    Args:
        sorted_ref_data (np.ndarray): Vertically sorted reference data.
        sorted_tgt_data (_type_): Vertically sorted perturbed data.
        use_continuity (bool, optional): Apply continuity factor or not . Defaults to True.
        alternative (Literal["two-sided", "less", "greater"]): Type of alternative hypothesis

    Returns:
        tuple[np.ndarray]: two-sided p-values, U-statistics. Each of shape (n_genes,).

    Author: Rémy Dubois
    """
    n_ref, ncols = sorted_ref_data.shape
    n_tgt, _ = sorted_tgt_data.shape

    U_statistics = np.empty(ncols, dtype=np.float64)
    pvals = np.empty(ncols, dtype=np.float64)
    n = n_ref + n_tgt
    mu = n_ref * n_tgt / 2.0
    for j in range(ncols):
        R1, tie_sum = rank_sum_and_ties_from_sorted(sorted_ref_data[:, j], sorted_tgt_data[:, j])

        # Compute U-stat
        U1 = n_ref * n_tgt + n_tgt * (n_tgt + 1) / 2 - R1

        pvals[j] = compute_pval(
            n_ref=n_ref,
            n_tgt=n_tgt,
            n=n,
            tie_sum=tie_sum,
            U=U1,
            mu=mu,
            contin_corr=0.5 if use_continuity else 0.0,
            alternative=alternative,
        )
        U_statistics[j] = U1

    return pvals, U_statistics


@njit(nogil=True, fastmath=True, cache=False, boundscheck=False)
def dense_ovo_mwu_kernel_over_contiguous_col_chunk(
    X: np.ndarray,
    chunk_lb: int,
    chunk_ub: int,
    grpc: GroupContainer,
    is_log1p: bool,
    use_continuity: bool = True,
    alternative: Literal["two-sided", "less", "greater"] = "two-sided",
) -> tuple[np.ndarray]:
    """Perform OVO tests group-wise and gene(col)-wise.

    Update: There is no need to fortranize the whole chunk at once, it can be done group by group within the loop. The memo

    Memory footprint investigations:
    1. ad.read_h5ad allocates memory in such a weird way that zeros are not properly assigned. Most likely this differs on unix systems. All investigations done on my MBP were so weird because of that.
    Tests were so weird because during the wilcoxon test, values were accessed and re-ordered and all those things. As a result, some memory allocation was happening within my functions although
    I was not allocating anything.
    2. np.asfortranarray(X[:, chunk_lb:chunk_ub][grpc.indices]) seems to allocate 2x the memory needed for the chunk. So if chunk is 2GB, this line allocates 4GB temporarily.
    3.


    Args:
        X (np.ndarray): Input dense expression matrix of shape (n_cells, n_genes)
        chunk_lb (int): Lower bound of the vertical slicing
        chunk_ub (int): Upper bound of the vertical slicing
        grpc (GroupContainer): GroupContainer, contains information about which group each row belongs to.
        use_continuity (bool, optional): Apply continuity factor or not. Defaults to True.
        is_log1p (bool, optional): User-indicated flag telling if data underwent log1p transform or not. Defaults to False.

    Raises:
        ValueError: If bounds are not intelligible.

    Returns:
        tuple[np.ndarray]: two-sided p-values, U-statistics, fold change. Each
        of shape (n_groups, chunk_ub - chunk_lb).

    Author: Rémy Dubois
    """
    chunk = X[:, chunk_lb:chunk_ub]
    n_groups = grpc.counts.size

    ref_indices = grpc.indices[grpc.indptr[grpc.encoded_ref_group] : grpc.indptr[grpc.encoded_ref_group + 1]]
    # TODO: still have to benchmark speedup of F order
    ref_chunk = chunk_and_fortranize(X, chunk_lb, chunk_ub, ref_indices)
    _sort_along_axis_inplace(ref_chunk, axis=0)

    pvalues = np.empty((n_groups, chunk_ub - chunk_lb), dtype=np.float64)
    statistics = np.empty((n_groups, chunk_ub - chunk_lb), dtype=np.float64)
    for group_id in range(n_groups):
        if group_id == grpc.encoded_ref_group:
            continue
        tgt_indices = grpc.indices[grpc.indptr[group_id] : grpc.indptr[group_id + 1]]
        # tgt_chunk = np.asfortranarray(chunk[tgt_indices, :])
        tgt_chunk = chunk_and_fortranize(X, chunk_lb, chunk_ub, tgt_indices)
        _sort_along_axis_inplace(tgt_chunk, axis=0)

        pvalues[group_id], statistics[group_id] = dense_ovo_mwu_kernel(
            sorted_ref_data=ref_chunk,
            sorted_tgt_data=tgt_chunk,
            use_continuity=use_continuity,
            alternative=alternative,
        )

    # Compute fold change
    fc = dense_fold_change(chunk, grpc, is_log1p=is_log1p)

    return pvalues, statistics, fc



import gc
import math

import anndata as ad
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from loguru import logger
from scipy import sparse
from tqdm.auto import tqdm

from illico.ovo import ovo_mwu_over_contiguous_col_chunk
from illico.ovr import ovr_mwu_over_col_contiguous_chunk
from illico.utils.compile import _precompile
from illico.utils.groups import encode_and_count_groups
from illico.utils.memory import _log_memory_usage
from illico.utils.ranking import check_indices_sorted_per_parcel

# from illico.utils.math import _warn_log1p

__all__ = ["asymptotic_wilcoxon"]


def asymptotic_wilcoxon(
    adata: ad.AnnData,
    is_log1p: bool,
    group_keys: str,
    reference: str | None = None,
    n_threads: int = 1,
    batch_size: int = 256,
    alternative: str = "two-sided",
    use_continuity: bool = True,
    layer: str | None = None,
    precompile: bool = True,
):
    """Perform asymptotic Mann-Whitney tests for differential gene expression.

    Mann-Whitney test is the same as Wilcoxon rank-sum test.

    This function takes as input an AnnData object of shape (n-cells, n-genes) with a group
    (e.g., perturbation) variable stored in .obs. It performs either one-versus-rest (OVR) or
    one-versus-one (OVO) Wilcoxon-Mann-Whitney tests for each gene, depending on whether a
    reference group is provided.

    Parameters
    ----------
    adata:
        Annotated data matrix of shape (n-cells, n-genes).
    is_log1p
        Whether the data is log1p transformed.
    group_keys
        Key in `adata.obs` specifying the group variable.
    reference
        Name of the reference group for OVO tests. If `None`, OVR tests are performed.
    n_threads
        Number of threads to use for parallel computation.
    batch_size
        Number of genes to process in each batch.
    alternative
        Type of alternative hypothesis. One of 'two-sided', 'less', or 'greater'.
    use_continuity
        Whether to apply continuity correction.
    layer
        Layer in `adata.layers` to use for the data. If `None`, uses `adata.X`.
    precompile
        Whether to precompile necessary functions for performance.

    Returns
    -------
    pd.DataFrame
        A DataFrame with MultiIndex (group, gene) containing p-values, statistics, and fold changes.

    Author: Rémy Dubois
    """
    # TODO: add a sparsity warning inviting user to convert to sparse if possible
    # TODO: rename group_keys

    if layer is not None:
        logger.info(f"Using layer '{layer}' for differential expression.")
        X = adata.layers[layer]
    else:
        X = adata.X

    # Check that the input CSR is sorted.
    if isinstance(X, sparse.csr_matrix):
        if not check_indices_sorted_per_parcel(X.indices, X.indptr):
            raise ValueError(
                "Input data matrix indices are not sorted. This is very unusual and may lead to incorrect results. "
                "This can be the result of operations like `adata[:, np.random.choice(…)]` that do not preserve sorting."
                "Please make sure that indices used to chunk the adata or the expression matrix have been sorted "
                "prior to computing DE genes."
            )

    if precompile:
        _precompile(X, reference)

    # Process the groups information
    raw_groups = adata.obs[group_keys].tolist()
    unique_raw_groups, group_container = encode_and_count_groups(groups=raw_groups, ref_group=reference)
    logger.info(
        f"Found {group_container.counts.size} unique groups (min size: {group_container.counts.min()} cells; max size: {group_container.counts.max()} cells), with reference group: {reference}"
    )
    _, n_genes = X.shape

    # Allocate the results dataframes
    cols = pd.Series(adata.var_names, name="feature", dtype=str)
    rows = pd.Series(unique_raw_groups, name="pert", dtype=str)
    results = np.empty((len(rows), len(cols), 3), dtype=np.float64)

    # Adapt batch size to leverage multithreading regarding the number of genes, if requested
    batch_size = min(batch_size, math.ceil(n_genes / n_threads))
    logger.trace(f"Using batch size of {batch_size} for {n_threads} threads and {n_genes} genes.")

    # Compute estimated mem footprint
    _log_memory_usage(adata, group_container, batch_size, n_threads)

    # Generate batches of columns
    bounds = np.append(np.arange(0, n_genes, batch_size), n_genes)
    iterator = list(zip(bounds[:-1], bounds[1:]))
    n_tests = n_genes * group_container.counts.size
    logger.trace(f"Performing a total of {n_tests:,d} tests.")
    with Parallel(n_threads, prefer="threads", return_as="generator_unordered") as pool:
        with tqdm(total=n_tests, smoothing=0.0, unit="it", unit_scale=True, unit_divisor=1000) as pbar:
            if reference is None:  # ovr use case
                pbar.set_description("Running one-versus-all MannWhitney-U tests")
                op = delayed(lambda *args: (ovr_mwu_over_col_contiguous_chunk(*args), args))
            else:  # ovo use case
                pbar.set_description("Running one-versus-ref MannWhitney-U tests")
                op = delayed(lambda *args: (ovo_mwu_over_contiguous_col_chunk(*args), args))
            for (pv, ustat, fc), args in pool(
                op(X, lb, ub, group_container, is_log1p, use_continuity, alternative) for lb, ub in iterator
            ):
                # progress.update(task, advance=group_container.counts.size * (args[2] - args[1]))  # refresh after processing is done
                pbar.update(group_container.counts.size * (args[2] - args[1]))
                col_chunk = slice(*args[1:3])
                results[:, col_chunk, 0] = pv
                results[:, col_chunk, 1] = ustat
                results[:, col_chunk, 2] = fc

                # Cleanup memory
                del pv, ustat, fc
                gc.collect()

        # Return a pd.DataFrame to index results
        results = pd.DataFrame(
            data=results.reshape(-1, 3),
            index=pd.MultiIndex.from_product([rows, cols], names=["pert", "feature"]),
            columns=["p_value", "statistic", "fold_change"],
        )

        """
            This case is harder than the OVR use case, because:
            1. If input is dense, then it's easy we can slice easily over rows (or group of rows) or columns and parallelize the way we want
            2. If input is CSR, we can easily slice the perturbations, and easily slice contiguous columns
            3. If input is CSC, we can easily slice along the columns, BUT NOT ALONG THE PERTURBATIONS because perturbations are non-contiguous along axis 0

            So the answer is simple: if we want a unified parallelism scheme, it has to be along the columns. Now:
            - For the dense use case, there is no problem.
            - For the sparse use case, the OVO kernel requires data to be CSC, so:
                a) If input is CSC, it means we do overall: big CSC ->  vertical chunk of CSC -> horizontal chunk of vertical chunk of CSC -> test
                b) If input is CSR, it means we do: big CSR  ->  vertical chunk of CSR -> horizontal chunk of vertical chunk of CSR -> CSC -> test
            For a), the second chunking is suboptimal because we chunk along 0 a CSC matrix, but because it has little columns this is fast.
            For b), this is a lot of conversion but I believe they all happen on small matrices so this should be fine.

            Conclusion: runtime on H1 is ~25 seconds for CSC, 25 seconds for CSR so still very nice.
            TO KEEP IN MIND: there is **no** way to optimize any other parallelism than column-based actually, the current idea (parallelizing over the
            perturbations) will be very slow, or require a brand new CSR allocation (which we don't want), to function.
            """
    return results
